import numpy as np
import hyperspy.api as hs
from hyperspy._signals.signal2d import Signal2D

from AmorphSim.utils.rotation_utils import _get_rotation_matrix, _get_random_2d_rot, _get_random_3d_rot
from AmorphSim.utils.simulation_utils import _get_speckle_size, _get_wavelength, _shape_function, _get_speckle_intensity
from AmorphSim.utils.vector_utils import rotation_matrix_from_vectors, build_ico
from skimage.draw import circle
from skimage.filters import gaussian
from numpy.random import random, choice
import matplotlib.pyplot as plt
from matplotlib.patches import Circle


class SimulationCube(object):
    """Defines a simulation cube of dimensions x, y, z in nm.  This allows you to create some simulation of the cube
    based on kinematic diffraction"""
    def __init__(self, dimensions=(20, 20, 20)):
        """Initializes the simulation cube
        Parameters
        --------------
        dimensions: tuple
            The dimensions of the cube to simulate.
        """
        self.clusters = []
        self.dimensions = np.array(dimensions)

    def __str__(self):
        return ("<Cube of " + str(self.num_clusters) + " clusters [" +
               str(self.dimensions[0])+" x "+str(self.dimensions[1])
               +" x "+str(self.dimensions[2])+"nm]>")

    @property
    def num_clusters(self):
        return len(self.clusters)

    def add_random_clusters(self, num_clusters, radius_range=(.5, 1.0), k_range=(3.5, 4.5), random_rotation=True,
                            symmetry=[2, 4, 6, 10]):
        """Randomly initializes the glass with a set of random clusters.
        Parameters
        ------------
        num_clusters: int
            The number of cluster to add
        radius_range: tuple
            The range of radii to randomly choose from
        k_range: tuple
            THe range of k to randomly choose from
        random_rotation: bool
            Random rotate the cluster or not
        symmetry: list
            The list of symmetries to choose from.  Acceptable symmetries are 2,4,6 and 10
        """
        rand_r = random(num_clusters) * (radius_range[1] - radius_range[0]) + radius_range[0]
        rand_k = random(num_clusters) * (k_range[1] - k_range[0]) + k_range[0]
        rand_sym = choice(symmetry,num_clusters)
        rand_pos = np.multiply(random((num_clusters,3)), self.dimensions)
        if random_rotation:
            rot_3d = _get_random_3d_rot(num_clusters)
            rot_2d = _get_random_2d_rot(num_clusters)
        else:
            rot_2d = [np.eye(3), ]*num_clusters
            rot_3d = [np.eye(3), ]*num_clusters
        for s, r, k, p, two, three in zip(rand_sym,rand_r, rand_k,rand_pos, rot_2d, rot_3d):
            self.clusters.append(Cluster(s,r,k,p,two,three))
        return

    def add_icoso(self, num_clusters, radius_range=(.5, 1.0), k_range=(3.5, 4.5), random_rotation=True):
        """Randomly initializes the glass with a set of random clusters.
        Parameters
        ------------
        num_clusters: int
            The number of cluster to add
        radius_range: tuple
            The range of radii to randomly choose from
        k_range: tuple
            THe range of k to randomly choose from
        random_rotation: bool
            Random rotate the cluster or not
        symmetry: list
            The list of symmetries to choose from.  Acceptable symmetries are 2,4,6 and 10
        """
        rand_r = random(num_clusters) * (radius_range[1] - radius_range[0]) + radius_range[0]
        rand_k = random(num_clusters) * (k_range[1] - k_range[0]) + k_range[0]
        rand_pos = np.multiply(random((num_clusters,3)), self.dimensions)
        if random_rotation:
            rot_3d = _get_random_3d_rot(num_clusters)
            rot_2d = _get_random_2d_rot(num_clusters)
        else:
            rot_2d = [np.eye(3), ]*num_clusters
            rot_3d = [np.eye(3), ]*num_clusters
        five_vertexes, three_face, two_edge = build_ico()
        for r, k, p, two, three in zip(rand_r, rand_k,rand_pos, rot_2d, rot_3d):
            for v in five_vertexes:
                self.clusters.append(Cluster(10,r,k,p,two,three,plane_direction=v))
            for v in three_face:
                self.clusters.append(Cluster(6,r,k,p,two,three,plane_direction=v))
            for v in two_edge:
                self.clusters.append(Cluster(2, r, k, p, two, three, plane_direction=v))
        return

    def show_projection(self, size=(512,512), acceptance = None):
        """Plots the 2-d projection.  Creates a 2-D projection of the clusters in the amorphous matrix.

        Parameters:
        ------------
        acceptance: float
            The angle of acceptance.  Only clusters within this projection will be allowed.
        size: int
            The size of the image made

        """
        projection = np.zeros(size)
        scale = [self.dimensions[0]/size[0],self.dimensions[1]/size[1]]
        for cluster in self.clusters:
            if acceptance is not None:
                if abs(cluster.get_angle_between()) > acceptance:
                    next()
            inten = np.sum(cluster.get_intensity())
            r, c = circle(cluster.position[0]/scale[0],
                          cluster.position[1]/scale[1],
                          radius=cluster.radius/scale[0],
                          shape=size)
            projection[r, c] = inten + projection[r, c]
        return projection

    def plot_symmetries(self, symmetries=[2, 4, 6, 8, 10], norm=1,acceptance=None):
        """Plots the 2-d projection of the symmetries as circles

        Parameters:
        ------------
        acceptance: float
            The angle of acceptance.  Only clusters within this projection will be allowed.
        size: int
            The size of the image made

        """
        fig, ax = plt.subplots()
        ax.set_xlim(0,self.dimensions[0])
        ax.set_ylim(0,self.dimensions[1])
        colors = ["black","blue","red","green","yellow","red","orange", "purple"]
        for cluster in self.clusters:
            if acceptance is not None:
                if abs(cluster.get_angle_between()) > acceptance:
                    continue
            inten = np.mean(cluster.get_intensity())
            c = Circle((cluster.position[0],
                        cluster.position[1]),
                        radius=cluster.radius,
                        alpha=inten/norm,
                        color=colors[symmetries.index(cluster.symmetry)])
            ax.add_patch(c)
        from matplotlib.lines import Line2D
        leg = [Line2D([0], [0], marker='o', color=colors[i], label=str(sym)+" fold symmetry",
               markerfacecolor=colors[i], markersize=15) for i,sym in enumerate(symmetries)]

        ax.legend(handles=leg)
        plt.show()
        return

    def get_4d_stem(self, convergence_angle=.74, accelerating_voltage=200,
                    k_rad = 5.0, simulation_size=(50, 50, 128, 128),
                    noise = False, disorder = None, num_electrons=1000, convolve=False, beam_size=None):
        """Returns an amorphous2d object which shows the 4d STEM projection for some set of clusters along some
        illumination

        Parameters
        ------------
        convergence_angle: float
            The convergance angle for the experiment
        accelerating_voltage: float
            The accelerating voltage for the experiment in kV
        simulation_size: tuple
            The size of the image for both the reciporical space image and the real space image.

        Returns
        ------------
        dataset: Amorphus2D
            Returns a 4 dimensional dataset which represents the cube
        """
        # dataset = Signal2D(np.ones(simulation_size))
        dataset = np.ones(simulation_size)
        real_scale = simulation_size[0] / self.dimensions[0]

        for cluster in self.clusters:
            speckles, observed_intensity = cluster.get_speckles(img_size=k_rad*2,
                                                                num_pixels=simulation_size[2],
                                                                accelerating_voltage=accelerating_voltage,
                                                                conv_angle=convergence_angle,
                                                                disorder=disorder)

            rr, rc = circle(r=cluster.position[0] * real_scale,
                            c=cluster.position[1] * real_scale,
                            radius=cluster.radius * real_scale,
                            shape=(simulation_size[0], simulation_size[1]))
            for (sr,sc), inten in zip(speckles, observed_intensity):
                inner_r, outer_r = np.meshgrid(sr, rr)
                inner_c, outer_c = np.meshgrid(sc, rc)
                if noise:
                    #print(np.size(inner_c))
                    inten = np.random.poisson(inten*num_electrons,size=np.size(inner_c))
                else:
                    inten= inten* num_electrons
                dataset[(outer_c.flatten(),
                         outer_r.flatten(),
                         inner_c.flatten(),
                         inner_r.flatten())] = inten + dataset[(outer_c.flatten(),
                                                                outer_r.flatten(),
                                                                inner_c.flatten(),
                                                                inner_r.flatten())]
        dataset = Signal2D(dataset)
        if convolve:
            if beam_size is None:
                beam_size = .9/convergence_angle  # rough calculation
            from scipy.signal import convolve2d
            num_pixels = int(real_scale*beam_size)
            xx,yy= np.ogrid[-num_pixels-1:num_pixels+2,-num_pixels-1:num_pixels+2]
            kernel =1 - ((xx ** 2 + yy ** 2) ** .5 - num_pixels).clip(0,1)
            dataset = dataset.T
            dataset.map(convolve2d, in2=kernel, mode="same", inplace=True)
            dataset = dataset.T


        dataset.axes_manager.navigation_axes[0].scale = self.dimensions[0]/simulation_size[0]
        dataset.axes_manager.navigation_axes[1].scale = self.dimensions[1] / simulation_size[1]
        dataset.axes_manager.navigation_axes[0].units = "nm"
        dataset.axes_manager.navigation_axes[1].units = "nm"
        dataset.axes_manager.signal_axes[0].scale = k_rad*2 /simulation_size[2]
        dataset.axes_manager.signal_axes[1].scale = k_rad*2 / simulation_size[3]
        dataset.axes_manager.signal_axes[0].units = "$nm^-1$"
        dataset.axes_manager.signal_axes[1].units = "$nm^-1$"
        dataset.axes_manager.signal_axes[0].offset = -k_rad
        dataset.axes_manager.signal_axes[1].offset = -k_rad
        return dataset


class Cluster(object):
    def __init__(self,
                 symmetry=10,
                 radius=1,
                 k=4.0,
                 position=random(2),
                 rotation_2d=np.eye(3),
                 rotation_3d=np.eye(3),
                 plane_direction=[0,0,1]):
        """Defines a cluster with a symmetry of symmetry, a radius of radius in nm and position of position.

        Parameters:
        ----------------
        symmetry: int
            The symmetry of the cluster being simulated
        radius: float
            The radius of the cluster in nm
        position: tuple
            The position of the cluster in the simulation cube
        rotation_vector: tuple
            The vector which the cluster is rotated around
        rotation_angle: float
            The angle the cluster is rotated about
        """
        self.symmetry = symmetry
        self.radius = radius
        self.position = position
        self.rotation_2d = np.eye(3)
        self.rotation_3d = rotation_3d
        self.k = k
        self.plane_direction = plane_direction
        self.beam_direction = [0,0,1]

    def get_diffraction(self, img_size=8.0,
                        num_pixels=512,
                        accelerating_voltage=200,
                        conv_angle=0.6,
                        disorder=None):
        """Takes some image size in inverse nm and then plots the resulting
        """
        sphere_radius = 1/_get_wavelength(accelerating_voltage)
        scale = (num_pixels-1)/img_size
        k_rotated=self.get_k_vectors()
        observed_intensity = [_get_speckle_intensity(k_vector=k,
                                                     ewald_sphere_rad=sphere_radius,
                                                     disorder=disorder,
                                                     cluster_rad=self.radius,
                                                     beam_direction=self.beam_direction)
                              for k in k_rotated]
        radius = _get_speckle_size(accelerating_voltage, conv_angle)*scale
        circles = [circle(int(k1[0] * scale + num_pixels/2), int(k1[1] * scale + num_pixels/2),
                          radius=radius) for k1 in k_rotated]
        image = np.ones(shape=(num_pixels,num_pixels))
        for (r, c), i in zip(circles, observed_intensity):
            image[r, c] = i + image[r, c]
        image = gaussian(image=image, sigma=2)
        return image

    def get_k_vectors(self):
        angle = (2 * np.pi) / self.symmetry  # angle between speckles on the pattern
        k = [[np.cos(angle * i) * self.k, np.sin(angle * i) * self.k, 0] for i in
             range(self.symmetry)]  # vectors for the speckles perp to BA
        #print(self.plane_direction)
        if not np.allclose(self.plane_direction, [0, 0, 1]) and not np.allclose(self.plane_direction, [0, 0, -1]) :
            rot = rotation_matrix_from_vectors([0,0,1],self.plane_direction)
            k = [np.dot(rot,v) for v in k]
        k_rotated2d = [np.dot(self.rotation_2d, speckle) for speckle in k]
        k_rotated = [np.dot(self.rotation_3d, speckle) for speckle in k_rotated2d]
        return k_rotated

    def get_speckles(self, img_size=10.0,
                     num_pixels=128,
                     accelerating_voltage=200,
                     conv_angle=0.6,
                     disorder=None):
        """
        This function returns the diffraction speckles as circles as defined by the
        skimage.draw.Circle class. Each speckle also has some intensity associated with it.

        Parameters
        -----------
        img_size:
            The real size of the image that is being projected onto. (In inverse nm)
        num_pixels:
            The pixelated size of the image being projected onto.
        accelerating_voltage:
            The accelerating voltage for getting the Ewald sphere radius
        conv_angle:
            The convergence angle in mrad for beam.
        """
        sphere_radius = 1/_get_wavelength(accelerating_voltage)
        scale = (num_pixels-1)/img_size
        k_rotated = self.get_k_vectors()
        observed_intensity = [_get_speckle_intensity(k_vector=k,
                                                     ewald_sphere_rad=sphere_radius,
                                                     disorder=disorder,
                                                     cluster_rad=self.radius,
                                                     beam_direction=self.beam_direction)
                              for k in k_rotated]
        radius = _get_speckle_size(accelerating_voltage, conv_angle)*scale
        speckles = [circle(int(k1[0] * scale + num_pixels/2), int(k1[1] * scale + num_pixels/2),
                           radius=radius, shape=(num_pixels,num_pixels)) for k1 in k_rotated]
        return speckles, observed_intensity

    def get_intensity(self,accelerating_voltage=200, disorder=None):
        """Takes some image size in inverse nm and then plots the resulting
        """
        sphere_radius = 1 / _get_wavelength(accelerating_voltage)
        k_rotated = self.get_k_vectors()
        observed_intensity = [_get_speckle_intensity(k_vector=k,
                                                     ewald_sphere_rad=sphere_radius,
                                                     disorder=disorder,
                                                     cluster_rad=self.radius,
                                                     beam_direction=self.beam_direction)
                              for k in k_rotated]
        return observed_intensity

    def get_angle_between(self):
        return np.arccos((np.trace(self.rotation_3d)-1)/2)



import numpy as np

def rotation_matrix_from_vectors(vec1, vec2):
    """ Find the rotation matrix that aligns vec1 to vec2
    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
    """
    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))
    return rotation_matrix

def build_ico():
    five_vertexes = np.array([[0,1,1.9], [0,-1,1.9],[0,1,-1.9], [0,-1,-1.9],
                              [1,1.9,0], [-1,1.9,0],[1,-1.9,0], [-1,-1.9,0],
                              [1.9,0,1], [-1.9,0,1],[1.9,0,-1], [-1.9,0,-1]])
    n = np.linalg.norm(five_vertexes, axis=1)
    five_vertexes = np.divide(five_vertexes, n[:,np.newaxis])
    two_edge = []
    for v1 in five_vertexes:
        for v2 in five_vertexes:
            #two_edge.append(np.linalg.norm(np.array(v1)-np.array(v2)))
            #print(np.linalg.norm(np.array(v1)-np.array(v2)))
            if np.abs(np.linalg.norm(np.array(v1)-np.array(v2)) - 1.08) <.3 :
               two_edge.append(v1+v2)
    n = np.linalg.norm(two_edge, axis=1)
    two_edge = np.divide(two_edge, n[:,np.newaxis])
    two_edge = np.unique(two_edge, axis=0)
    three_face = []
    for v1 in five_vertexes:
        for v2 in five_vertexes:
            for v3 in five_vertexes:
                n1 = np.abs(np.linalg.norm(np.array(v1) - np.array(v2))-1.08)
                n2 = np.abs(np.linalg.norm(np.array(v1) - np.array(v3))-1.08)
                n3 = np.abs(np.linalg.norm(np.array(v2) - np.array(v3))-1.08)
                if  n1 < .3 and n2 < .3 and  n3 < .3:
                    three_face.append(v1+v2+v3)
    n = np.linalg.norm(three_face, axis=1)
    three_face = np.divide(three_face, n[:,np.newaxis])
    three_face = np.unique(three_face,axis=0)
    return five_vertexes, three_face, two_edge


import numpy as np
from numpy.random import random

def _get_random_rotation(num_vectors):
    rand_vector = random((num_vectors, 3)) * 2 - 1  # change
    norms = np.linalg.norm(rand_vector, axis=1)
    rand_vector = np.divide(rand_vector,norms[:,np.newaxis])
    rand_rot = random(num_vectors) * np.pi
    return rand_vector, rand_rot

def _rand_2d_rotation_matrix():
    angle = np.random.uniform(low=0.0, high=np.pi)
    M = np.array([[np.cos(angle),-np.sin(angle), 0],
                  [np.sin(angle), np.cos(angle), 0],
                  [0,0,1]])
    return M


def _rand_3d_rotation_matrix(deflection=1.0, randnums=None):
    """
    Creates a random rotation matrix.

    deflection: the magnitude of the rotation. For 0, no rotation; for 1, competely random
    rotation. Small deflection => small perturbation.
    randnums: 3 random numbers in the range [0, 1]. If `None`, they will be auto-generated.
    """
    # from http://www.realtimerendering.com/resources/GraphicsGems/gemsiii/rand_rotation.c
    randnums = np.random.uniform(size=(3,))
    theta, phi, z = randnums
    theta = theta * 2.0 * deflection * np.pi  # Rotation about the pole (Z).
    phi = phi * 2.0 * np.pi  # For direction of pole deflection.
    z = z * 2.0 * deflection  # For magnitude of pole deflection.
    r = np.sqrt(z)
    Vx, Vy, Vz = V = (np.sin(phi) * r, np.cos(phi) * r, np.sqrt(2.0 - z))
    st = np.sin(theta)
    ct = np.cos(theta)
    R = np.array(((ct, st, 0), (-st, ct, 0), (0, 0, 1)))
    # Construct the rotation matrix  ( V Transpose(V) - I ) R.
    M = (np.outer(V, V) - np.eye(3)).dot(R)
    return M


def _get_random_2d_rot(num_vectors):
    rot_m = []
    for i in range(0, num_vectors):
        M = _rand_2d_rotation_matrix()
        rot_m.append(M)
    return rot_m


def _get_random_3d_rot(num_vectors):
    rot_m = []
    for i in range(0, num_vectors):
        M = _rand_3d_rotation_matrix()
        rot_m.append(M)
    return rot_m


def _get_rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians. Using Euler-Rodrigues Formula
    """
    axis = np.asarray(axis)
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(theta / 2.0)
    b, c, d = -axis * np.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])






from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='AmorphSim',
      version='0.02',
      description='Tool for simulating amorphous Materials',
      long_description=readme(),
      keywords='Simulations STEM Electron Microscopy Glass',
      url='https://github.com/CSSFrancis/AmorphSim',
      author='CSSFrancis',
      author_email='csfrancis@wisc.edu',
      liscense='MIT',
      packages=['AmorphSim',
                'AmorphSim.utils',],
      install_requires=['hyperspy >=1.5',
                        'numpy>=1.10,!=1.70.0',
                        'matplotlib',
                        'scipy',
                        'scikit-image'],
      zip_safe=False)



class GlassCluster(object):
    def __init__(self, vertexes=None, radius=1, k=4.0, position=random(2),
                 rotation_vector=[1, 0, 0], rotation_angle=0, diffraction_intensity=600):
        """Defines a cluster with a symmetry of symmetry, a radius of radius in nm and position of position.

        Parameters:
        ----------------
        symmetry: int
            The symmetry of the cluster being simulated
        radius: float
            The radius of the cluster in nm
        position: tuple
            The position of the cluster in the simulation cube
        rotation_vector: tuple
            The vector which the cluster is rotated around
        rotation_angle: float
            The angle the cluster is rotated about
        """
        self.symmetry = symmetry
        self.radius = radius
        self.position = position
        self.rotation_vector = rotation_vector
        self.rotation_angle = rotation_angle
        self.k = k
        self.diffraction_intensity = diffraction_intensity

    def get_diffraction(self, img_size=8.0, num_pixels=512, accelerating_voltage=200, conv_angle=0.6):
        """Takes some image size in inverse nm and then plots the resulting
        """
        rotation_matrix = _get_rotation_matrix(self.rotation_vector, self.rotation_angle)
        sphere_radius = 1/_get_wavelength(accelerating_voltage)
        scale = (num_pixels-1)/img_size
        angle = (2 * np.pi) / self.symmetry  # angle between speckles on the pattern
        k = [[np.cos(angle * i) * self.k, np.sin(angle * i) * self.k,0] for i in
             range(self.symmetry)]  # vectors for the speckles perp to BA
        k_rotated = [np.dot(rotation_matrix, speckle) for speckle in k]
        deviation = [_get_deviation(sphere_radius,speckle) for speckle in k_rotated]
        observed_intensity = [self.diffraction_intensity/self.symmetry * _shape_function(radius=self.radius, deviation=dev)
                              for dev in deviation]
        radius = _get_speckle_size(accelerating_voltage, conv_angle)*scale
        circles = [circle(int(k1[0] * scale + num_pixels/2), int(k1[1] * scale + num_pixels/2),
                          radius=radius) for k1 in k_rotated]
        image = np.ones(shape=(num_pixels,num_pixels))
        for (r, c), i in zip(circles, observed_intensity):
            image[r, c] = i + image[r, c]
        image = gaussian(image=image, sigma=2)
        return image

    def get_speckles(self, img_size=8.0, num_pixels=512, accelerating_voltage=200, conv_angle=0.6):
        """Takes some image size in inverse nm and then plots the resulting
        """
        rotation_matrix = _get_rotation_matrix(self.rotation_vector, self.rotation_angle)
        sphere_radius = 1/_get_wavelength(accelerating_voltage)
        scale = (num_pixels-1)/img_size
        angle = (2 * np.pi) / self.symmetry  # angle between speckles on the pattern
        k = [[np.cos(angle * i) * self.k, np.sin(angle * i) * self.k,0] for i in
             range(self.symmetry)]  # vectors for the speckles perp to BA
        k_rotated = [np.dot(rotation_matrix, speckle) for speckle in k]
        deviation = [_get_deviation(sphere_radius,speckle) for speckle in k_rotated]
        observed_intensity = [self.diffraction_intensity/self.symmetry * _shape_function(radius=self.radius, deviation=dev)
                              for dev in deviation]
        radius = _get_speckle_size(accelerating_voltage, conv_angle)*scale
        speckles = [circle(int(k1[0] * scale + num_pixels/2), int(k1[1] * scale + num_pixels/2),
                           radius=radius) for k1 in k_rotated]
        return speckles, observed_intensity

    def get_intensity(self,accelerating_voltage=200):
        """Takes some image size in inverse nm and then plots the resulting
        """
        rotation_matrix = _get_rotation_matrix(self.rotation_vector, self.rotation_angle)
        sphere_radius = 1/_get_wavelength(accelerating_voltage)
        angle = (2 * np.pi) / self.symmetry  # angle between speckles on the pattern
        k = [[np.cos(angle * i) * self.k, np.sin(angle * i) * self.k,0] for i in
             range(self.symmetry)]  # vectors for the speckles perp to BA
        k_rotated = [np.dot(rotation_matrix, speckle) for speckle in k]
        deviation = [_get_deviation(sphere_radius,speckle) for speckle in k_rotated]
        observed_intensity = [self.diffraction_intensity/self.symmetry * _shape_function(radius=self.radius, deviation=dev)
                              for dev in deviation]
        return observed_intensity

    def get_mistilt(self, v2=[1,1,1]):
        v1 = self.rotation_vector
        print(v1)
        return np.cos((v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])/(np.linalg.norm(v1)*np.linalg.norm(v2)))


import numpy as np
import hyperspy.api as hs


def _get_speckle_intensity(k_vector,
                           ewald_sphere_rad,
                           cluster_rad = 1.0,
                           disorder=None,
                           beam_direction=[0,0,1]):
    """Returns the intensity of some speckle out of one

    Parameters
    --------------
    k_vector:
        The vector for one of the diffraction spots
    ewald_sphere_rad:
        The radius of the Ewald Sphere.  This is equal to 1/wavelength
    disorder: None or float
        The amount of disorder in the glass.  This is a measure of the average
        displacement of some atom from ideality.
    :returns
    -----------
    observed_intensity:
        The intensity for some diffraction vector
    :return:
    """
    deviation = _get_deviation(ewald_sphere_rad,
                               k_vector,
                               beam_direction=beam_direction)
    observed_intensity =_shape_function(radius=cluster_rad,
                                        deviation=deviation)
    if disorder is not None:
        factor = _get_disorder(k_vector,disorder)
        observed_intensity = observed_intensity*factor
    return observed_intensity

def _shape_function(radius, deviation, function=None):
    """Returns the point at some deviation parameter s and a radius r

    Parameters
    ----------------
    r: float
        The radius of the cluster
    s_g: float
        The deviation parameter for the point.
    function: None
        The shape function to simulate (still need to add different functions)

    Returns:
    n:float
        Shape function for some deviation parameter and radius
    """
    c = 2* np.pi*deviation*radius
    n = ((3*(np.sin(c)-c*np.cos(c)))/c**3)**2
    return n


def s_g_kernel(kernel_size, d_hkl, cluster_size, voltage):
    """ Simulates a 2-d projection of the s_g kernel... (Maybe make 3-D if useful)
    Parameters
    ----------------
    kernel_size: int
        The size of the kernel to create
    d_hkl: float
        The interplanar spacing in n,
    cluster_size: float
        The size of the cluster being calculated.
    """
    wavelength = _get_wavelength(acc_voltage=voltage)
    ax = np.arange(-kernel_size // 2 + 1., kernel_size // 2 + 1.)
    xx, yy = np.divide(np.meshgrid(ax, ax), kernel_size / 3 * cluster_size)
    scaling = 1/(kernel_size / 3 * cluster_size)
    sg = np.power((np.square(xx) + np.square(yy)), 0.5)
    print(d_hkl)
    dot = np.subtract(2*wavelength*d_hkl, np.multiply(sg, d_hkl))
    angles = np.divide(sg, d_hkl)
    sg_surf = np.multiply(sg, (2 * np.pi * cluster_size))
    kernel = np.power(np.multiply(np.divide((np.sin(sg_surf) -
                                             np.multiply(sg_surf,
                                                         np.cos(sg_surf))),
                                            np.power(sg_surf, 3)), 3), 2)
    dict0 = {'size': kernel_size, 'name': 's_x', 'units': 'nm^-1', 'scale': scaling, 'offset': 0}
    dict1 = {'size': kernel_size, 'name': 's_y', 'units': 'nm^-1', 'scale': scaling, 'offset': 0}
    k = hs.signals.Signal2D(data=kernel, axes=[dict0, dict1])
    return k

def s_g_kernel_angle(kernel_size, d_hkl, cluster_size, voltage):
    """ Simulates a 2-d projection of the s_g kernel... (Maybe make 3-D if useful)
    Parameters
    ----------------
    kernel_size: int
        The size of the kernel to create
    d_hkl: float
        The interplanar spacing in n,
    cluster_size: float
        The size of the cluster being calculated.
    """
    wavelength = _get_wavelength(acc_voltage=voltage)
    ax = np.arange(-kernel_size // 2 + 1., kernel_size // 2 + 1.)
    xx, yy = np.divide(np.meshgrid(ax, ax), kernel_size / 3 * cluster_size)
    scaling = 1/(kernel_size / 3 * cluster_size)
    sg = np.power((np.square(xx) + np.square(yy)), 0.5)
    print(d_hkl)
    dot = np.subtract(2*wavelength*d_hkl, np.multiply(sg, d_hkl))
    angles = np.divide(sg, d_hkl)
    sg_surf = np.multiply(sg, (2 * np.pi * cluster_size))
    kernel = np.power(np.multiply(np.divide((np.sin(sg_surf) -
                                             np.multiply(sg_surf,
                                                         np.cos(sg_surf))),
                                            np.power(sg_surf, 3)), 3), 2)
    dict0 = {'size': kernel_size, 'name': 's_x', 'units': 'nm^-1', 'scale': scaling, 'offset': 0}
    dict1 = {'size': kernel_size, 'name': 's_y', 'units': 'nm^-1', 'scale': scaling, 'offset': 0}
    k = hs.signals.Signal2D(data=kernel, axes=[dict0, dict1])
    return k


def _get_speckle_size(accelerating_voltage=200, semi_angle= 0.74):
    """Gets the size of a speckle from the semi angle
    Parameters
    ----------------
    accelerating_voltage:
        The accelerating voltage for the microscope
    semi_angle: float
        The angle in mili-radians of the convergance
    k: float
        The k-vector length in nm
    """
    wavelength = _get_wavelength(acc_voltage=accelerating_voltage)
    size = np.sin(semi_angle/1000) * 1/wavelength
    return size


def _get_wavelength(acc_voltage):
    """Given some accelerating voltage for a microscope calculate the relativistic wavelength
    Parameters
    -----------------
    acc_voltage: float
        The accelerating voltage of the microscope.

    Returns:
    -----------
    wavelength:float
        The wavelength of the electrons.
    """
    h = 6.626*10**-34
    m0 = 9.109*10**-31
    e = 1.602*10**-19
    wavelength = h/np.sqrt(2*m0*acc_voltage*1000*e*(1+acc_voltage/(2*511)))*10**9
    return wavelength


def _get_deviation(sphere_radius, k, beam_direction=[0,0,-1]):
    """
    Parameters
    ----------------
    sphere_radius: float
        The radius of the sphere
    k0: tuple
        The (x,y,z) of the original s value from the optic axis.
    """
    beam = np.array(beam_direction) * sphere_radius
    dist = np.linalg.norm(beam-k)
    deviation = sphere_radius-dist # distance from edge of sphere to k
    return deviation

def _get_disorder(k, displacement):
    """The estimates the amount of disorder in the perfect crystal.
    This might need some additional work.

    Parameters
    ----------------
    sphere_radius: float
        The radius of the sphere
    k: tuple
        The (x,y,z) of the original s value from the optic axis.
    """
    b = 8 * np.pi**2 * displacement**2
    s = np.linalg.norm(k)
    factor = np.exp(-b*s**2)
    return factor